
#define OBJ_CIX         512

#define CLIP_XY         8
#define CLIP_Z          0.1
#define CLIP_FAR_Z      40
#define CLIP_LIT_Z      20
#define CLIP_TINY       4

#define TEX_F           32

#define COL_DUAL        0x01000000
#define COL_TRANS       0x02000000


typedef struct coords_struct {
   float x, y, z;
} coords_info;

typedef struct tile_struct {
   float u, v;
} tile_info;


typedef struct face_struct {
   struct face_struct *next;
   int       flags, ncix;
   BITMAP    *texture;
   int       *icix;
   tile_info *tix;
} face_info;

#define fi_f_tex        1
#define fi_f_mask       2
#define fi_f_atex       4
#define fi_f_cix        128


typedef struct surf_struct {
   struct surf_struct *next;
   int       flags, ncix, *icix, color, *colox;
   face_info *fix;
} surf_info;

#define si_f_dual       1
#define si_f_grgb       2
#define si_f_cix        128


typedef struct obj_struct {
   struct obj_struct *next;
   int         flags, ncix;
   coords_info *cix;
   surf_info   *six;
} obj_info;

#define oi_f_tiny       1
#define oi_f_moved      0x100


typedef struct light_struct {
   struct light_struct *next;
   int   flags, l;
   float x, y, z;
} light_info;

#define li_f_moved      0x1700


/* convert radians to degrees */
#define DEG(n)    ((n) * 180.0 / M_PI)

/* convert rgb color to hardware color */
#define rgb2r(c) (((c) >> 16) & 255)
#define rgb2g(c) (((c) >> 8) & 255)
#define rgb2b(c) ((c) & 255)
#define rgb2h(c) makecol(rgb2r(c), rgb2g(c), rgb2b(c))

/* convert hardware color to rgb */
#define h2rgb(c) ((getr(c) << 16) | (getg(c) << 8) | getb(c))

/* clip r,g,b values */
#define clip_c(c) { if (c < 0) c = 0; if (c > 255) c = 255; }


void *valloc(int n, int s);


#define LID_NONE        0
#define LID_BOT         1
#define LID_TOP         2
#define LID_ALL         3

extern int ordered_cix[];
extern obj_info *oix;
extern light_info *lix, *headlight;

void init_objs();


BITMAP *tex_granite(int size, int rgb, int d, int f);
BITMAP *tex_brick(int size, int rgb, int d, int f);

