
#include <math.h>

#include "allegro.h"
#include "g3d.h"


int prism_cix[][4] = {
   {0, 4, 5, 1},
   {1, 5, 6, 2},
   {2, 6, 7, 3},
   {3, 7, 4, 0},
   {0, 1, 2, 3},
   {7, 6, 5, 4}
};



int power2(int s)
{
   if (s >= 32) return (s + 31) & 0xffffffe0;
   if (s >= 16) return (s + 15) & 0xfffffff0;
   if (s >= 8) return (s + 7) & 0xfffffff8;
   return (s + 4) & 0xfffffffc;
}



void free_obj(obj_info *oi)
{
   surf_info *si, *sj;
   face_info *fi, *fj;

   for (si=oi->six; si; si=sj) {
      for (fi=si->fix; fi; fi=fj) {
         if (fi->flags & fi_f_cix) free(fi->icix);
         if (fi->flags & fi_f_tex) free(fi->tix);
         fj = fi->next;
         free(fi);
      }
      if (si->flags & si_f_cix) free(si->icix);
      if (si->flags & si_f_grgb) free(si->colox);
      sj = si->next;
      free(si);
   }
   free(oi->cix);
   free(oi);
}



obj_info *cylinder_v(float x, float y, float z, float r, float h,
                     int n, int lid, int color, BITMAP *tex)
{
   int i;
   double a = 0.0, ai = 2.0 * M_PI / n;
   int sg = power2((int)(TEX_F * r * sqrt(2.0 * (1.0 - cos(ai)))));
   obj_info *oi;
   surf_info *si, *ss, *sb = NULL, *st = NULL;
   face_info *fi;
   tile_info *ti;

   oi = valloc(1, sizeof(obj_info));
   oi->flags = oi_f_moved;
   oi->cix = valloc(2*n, sizeof(coords_info));
   oi->six = si = valloc(1, sizeof(surf_info));
   oi->ncix = si->ncix = 2*n;
   si->icix = valloc(2*n, sizeof(int));
   si->flags = si_f_cix;
   si->color = color;

   if (lid & LID_BOT) {
      si->next = sb = valloc(1, sizeof(surf_info));
      sb->ncix = n;
      sb->icix = si->icix;
      sb->color = color;
      sb->fix = fi = valloc(1, sizeof(face_info));
      fi->ncix = n;
      fi->icix = sb->icix;
      if (tex) {
         fi->flags = fi_f_tex;
         fi->texture = tex;
         fi->tix = valloc(n, sizeof(tile_info));
      }
   }
   if (lid & LID_TOP) {
      ss = si->next;
      si->next = st = valloc(1, sizeof(surf_info));
      st->next = ss;
      st->ncix = n;
      st->icix = valloc(n, sizeof(int));
      st->flags = si_f_cix;
      st->color = color;
      st->fix = fi = valloc(1, sizeof(face_info));
      fi->ncix = n;
      fi->icix = st->icix;
      if (tex) {
         fi->flags = fi_f_tex;
         fi->texture = tex;
         fi->tix = valloc(n, sizeof(tile_info));
      }
   }

   for (i=0; i<n; i++, a+=ai) {
      coords_info *cl = &oi->cix[i];
      coords_info *ch = &oi->cix[i+n];
      fi = valloc(1, sizeof(face_info));

      cl->x = ch->x = x + r * cos(a);
      cl->z = ch->z = z + r * sin(a);
      cl->y = y; ch->y = y - h;

      si->icix[i] = i; si->icix[i+n] = i+n;
      fi->next = si->fix; si->fix = fi;
      fi->ncix = 4;
      fi->icix = valloc(4, sizeof(int));
      fi->flags = fi_f_cix;
      fi->icix[0] = i; fi->icix[1] = i+n;
      if (i == n-1) {
         fi->icix[2] = n; fi->icix[3] = 0;
      } else {
         fi->icix[2] = i+n+1; fi->icix[3] = i+1;
      }
      if (lid & LID_TOP) {
         st->icix[i] = 2*n-i-1;
      }

      if (tex) {
         fi->flags = fi_f_tex;
         fi->texture = tex;
         fi->tix = ti = valloc(4, sizeof(tile_info));
         ti[0].u = ti[1].u = ti[1].v = ti[2].v = 0;
         ti[2].u = ti[3].u = sg;
         ti[0].v = ti[3].v = TEX_F * h;
         if (lid & LID_BOT) {
            ti = sb->fix->tix;
            ti[i].u = TEX_F * oi->cix[sb->icix[i]].x;
            ti[i].v = TEX_F * oi->cix[sb->icix[i]].z;
         }
         if (lid & LID_TOP) {
            ti = st->fix->tix;
            ti[i].u = TEX_F * oi->cix[st->icix[i]].x;
            ti[i].v = TEX_F * oi->cix[st->icix[i]].z;
         }
      }
   }

   return oi;
}



obj_info *cone_v(float x, float y, float z, float r, float h,
                 int n, int lid, int color, BITMAP *tex)
{
   int i;
   double a = 0.0, ai = 2 * M_PI / n;
   int sg = power2((int)(TEX_F * r * sqrt(2.0 * (1.0 - cos(ai)))));
   obj_info *oi;
   surf_info *si, *sb = NULL;
   face_info *fi;
   tile_info *ti;

   oi = valloc(1, sizeof(obj_info));
   oi->flags = oi_f_moved;
   oi->cix = valloc(2*n, sizeof(coords_info));
   oi->six = si = valloc(1, sizeof(surf_info));
   oi->ncix = si->ncix = 2*n;
   si->icix = valloc(2*n, sizeof(int));
   si->flags = si_f_cix;
   si->color = color;

   if (lid & LID_BOT) {
      si->next = sb = valloc(1, sizeof(surf_info));
      sb->ncix = n;
      sb->icix = si->icix;
      sb->color = color;
      sb->fix = fi = valloc(1, sizeof(face_info));
      fi->ncix = n;
      fi->icix = sb->icix;
      if (tex) {
         fi->flags = fi_f_tex;
         fi->texture = tex;
         fi->tix = valloc(n, sizeof(tile_info));
      }
   }

   for (i=0; i<n; i++, a+=ai) {
      coords_info *cl = &oi->cix[i];
      coords_info *ch = &oi->cix[i+n];
      fi = valloc(1, sizeof(face_info));

      cl->x = x + r * cos(a); ch->x = x;
      cl->z = z + r * sin(a); ch->z = z;
      cl->y = y; ch->y = y - h;

      si->icix[i] = i; si->icix[i+n] = i+n;
      fi->next = si->fix; si->fix = fi;
      fi->ncix = 3;
      fi->icix = valloc(3, sizeof(int));
      fi->flags = fi_f_cix;
      fi->icix[0] = i; fi->icix[1] = i+n;
      if (i == n-1) {
         fi->icix[2] = 0;
      } else {
         fi->icix[2] = i+1;
      }

      if (tex) {
         fi->flags = fi_f_tex;
         fi->texture = tex;
         fi->tix = ti = valloc(4, sizeof(tile_info));
         ti[0].u = ti[1].v = 0;
         ti[1].u = sg / 2; ti[2].u = sg;
         ti[0].v = ti[3].v = TEX_F * h;
         if (lid & LID_BOT) {
            ti = sb->fix->tix;
            ti[i].u = TEX_F * oi->cix[sb->icix[i]].x;
            ti[i].v = TEX_F * oi->cix[sb->icix[i]].z;
         }
      }
   }

   return oi;
}



obj_info *prism_v(float x, float y, float z, float w, float h, float d,
                  int lid, int color, BITMAP *tex)
{
   int i;
   obj_info *oi;
   surf_info *si;
   face_info *fi;
   tile_info *ti;

   oi = valloc(1, sizeof(obj_info));
   oi->flags = oi_f_moved;
   oi->ncix = 8;
   oi->cix = valloc(8, sizeof(coords_info));
   oi->cix[0].x = oi->cix[4].x = oi->cix[3].x = oi->cix[7].x = x + w / 2;
   oi->cix[1].x = oi->cix[5].x = oi->cix[2].x = oi->cix[6].x = x - w / 2;
   oi->cix[0].y = oi->cix[1].y = oi->cix[2].y = oi->cix[3].y = y;
   oi->cix[4].y = oi->cix[5].y = oi->cix[6].y = oi->cix[7].y = y - h;
   oi->cix[0].z = oi->cix[4].z = oi->cix[1].z = oi->cix[5].z = z + d / 2;
   oi->cix[2].z = oi->cix[6].z = oi->cix[3].z = oi->cix[7].z = z - d / 2;

   for (i=0; i<6; i++) {
      if ((i == 4) && !(lid & LID_BOT)) continue;
      if ((i == 5) && !(lid & LID_TOP)) continue;
      si = valloc(1, sizeof(surf_info));
      si->next = oi->six; oi->six = si;
      si->fix = fi = valloc(1, sizeof(face_info));
      si->ncix = fi->ncix = 4;
      si->icix = fi->icix = prism_cix[i];
      si->color = color;
      if (tex) {
         fi->flags = fi_f_tex;
         fi->texture = tex;
         fi->tix = ti = valloc(4, sizeof(tile_info));
         ti[0].u = ti[1].u = ti[1].v = ti[2].v = 0;
         if (i == 1 || i == 3) {
            ti[2].u = ti[3].u = TEX_F * d;
         } else {
            ti[2].u = ti[3].u = TEX_F * w;
         }
         if (i == 4 || i == 5) {
            ti[0].v = ti[3].v = TEX_F * d;
         } else {
            ti[0].v = ti[3].v = TEX_F * h;
         }
      }
   }

   return oi;
}



obj_info *face_v(float x1, float y1, float z1, float h1,
                 float x2, float y2, float z2, float h2,
                 int color, BITMAP *tex)
{
   obj_info *oi;
   surf_info *si;
   face_info *fi;
   tile_info *ti;

   oi = valloc(1, sizeof(obj_info));
   oi->flags = oi_f_moved;
   oi->ncix = 4;
   oi->cix = valloc(4, sizeof(coords_info));
   oi->cix[0].x = oi->cix[1].x = x1;
   oi->cix[2].x = oi->cix[3].x = x2;
   oi->cix[0].z = oi->cix[1].z = z1;
   oi->cix[2].z = oi->cix[3].z = z2;
   oi->cix[0].y = y1; oi->cix[1].y = y1 - h1;
   oi->cix[2].y = y2 - h2; oi->cix[3].y = y2;

   oi->six = si = valloc(1, sizeof(surf_info));
   si->fix = fi = valloc(1, sizeof(face_info));
   si->ncix = fi->ncix = 4;
   si->icix = fi->icix = ordered_cix;
   si->color = color;
   if (color & COL_DUAL) si->flags |= si_f_dual;
   if (tex) {
      float dx = x2 - x1, dz = z2 - z1;
      fi->flags = fi_f_tex;
      fi->texture = tex;
      fi->tix = ti = valloc(4, sizeof(tile_info));
      ti[0].u = ti[1].u = 0;
      ti[2].u = ti[3].u = TEX_F * sqrt(dx * dx + dz * dz);
      ti[0].v = TEX_F * h1;
      ti[1].v = 0;
      ti[2].v = TEX_F * (y2 - h2 - y1 + h1);
      ti[3].v = TEX_F * (y2 - y1 + h1);
   }

   return oi;
}



obj_info *rect_h(float x, float y, float z, float w, float d,
                 int color, BITMAP *tex)
{
   obj_info *oi;
   surf_info *si;
   face_info *fi;
   tile_info *ti;

   oi = valloc(1, sizeof(obj_info));
   oi->flags = oi_f_moved;
   oi->ncix = 4;
   oi->cix = valloc(4, sizeof(coords_info));
   oi->cix[0].x = oi->cix[1].x = x + w / 2;
   oi->cix[2].x = oi->cix[3].x = x - w / 2;
   oi->cix[0].z = oi->cix[3].z = z + d / 2;
   oi->cix[1].z = oi->cix[2].z = z - d / 2;
   oi->cix[0].y = oi->cix[1].y = oi->cix[2].y = oi->cix[3].y = y;

   oi->six = si = valloc(1, sizeof(surf_info));
   si->fix = fi = valloc(1, sizeof(face_info));
   si->ncix = fi->ncix = 4;
   si->icix = fi->icix = ordered_cix;
   si->color = color;
   if (color & COL_DUAL) si->flags |= si_f_dual;
   if (tex) {
      fi->flags = fi_f_tex;
      fi->texture = tex;
      fi->tix = ti = valloc(4, sizeof(tile_info));
      ti[0].u = ti[1].u = 0;
      ti[2].u = ti[3].u = TEX_F * w;
      ti[1].v = ti[2].v = 0;
      ti[0].v = ti[3].v = TEX_F * d;
   }

   return oi;
}



void init_objs()
{
   obj_info *oi;
   light_info *li;

   oi = cylinder_v(0, 0, 0, 1, 3, 16, LID_ALL, 0x00ffff, NULL);
   oi->next = oix; oix = oi;

   oi = cone_v(4, 0, 0, 1, 3, 16, LID_ALL, 0xffff00, NULL);
   oi->next = oix; oix = oi;

   oi = prism_v(-4, 0, 0, 2, 3, 2, LID_ALL, 0xff00ff, NULL);
   oi->next = oix; oix = oi;

   oi = cylinder_v(7, 2, 11, 0.5, 2, 8, LID_NONE, 0xff0000, NULL);
   oi->next = oix; oix = oi;
   oi = cone_v(7, 0, 11, 0.5, 1, 8, LID_NONE, 0xff8000, NULL);
   oi->next = oix; oix = oi;

   oi = face_v(-8, 2, -4, 8, -8, 2, 12, 8, COL_DUAL, tex_brick(16, 0x802050, 31, 100));
   oi->next = oix; oix = oi;
   oi = face_v(-8, 2, 12, 8, 8, 2, 12, 8, COL_DUAL, tex_granite(32, 0xc0c0c0, 31, 100));
   oi->next = oix; oix = oi;
   oi = face_v(8, 2, 12, 8, 8, 2, -4, 8, COL_DUAL, tex_granite(32, 0x808080, 63, 200));
   oi->next = oix; oix = oi;
   oi = face_v(0, 2, -4, 8, 8, 2, -4, 8, 0x0080ff | COL_DUAL | COL_TRANS, NULL);
   oi->next = oix; oix = oi;

   oi = rect_h(0, 2, 4, 18, 18, 0xc0c0c0, NULL);
   oi->next = oix; oix = oi;
   oi = rect_h(0, -6, 4, -18, 18, 0xc0c0c0, NULL);
   oi->next = oix; oix = oi;

   li = valloc(1, sizeof(light_info));
   li->next = lix; lix = li;
   li->l = 1000; li->x = 2; li->y = 1; li->z = -3;

   headlight = li = valloc(1, sizeof(light_info));
   li->next = lix; lix = li;
   li->l = 2000;
}

